<?php

namespace Drupal\access_records\EventSubscriber;

use Drupal\access_records\AccessRecordQueryBuilder;
use Drupal\access_records\AccessRecordTypeInterface;
use Drupal\access_records\Entity\AccessRecordType;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\entity\QueryAccess\Condition;
use Drupal\entity\QueryAccess\ConditionGroup;
use Drupal\entity\QueryAccess\QueryAccessEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Subscriber for query access events to include access records.
 */
class AccessRecordsQueryAccessSubscriber implements EventSubscriberInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The access record query builder.
   *
   * @var \Drupal\access_records\AccessRecordQueryBuilder
   */
  protected AccessRecordQueryBuilder $accessRecordQueryBuilder;

  /**
   * The AccessRecordsQueryAccessSubscriber constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $etm
   *   The entity type manager.
   * @param \Drupal\access_records\AccessRecordQueryBuilder $aqb
   *   The access record query builder.
   */
  public function __construct(EntityTypeManagerInterface $etm, AccessRecordQueryBuilder $aqb) {
    $this->entityTypeManager = $etm;
    $this->accessRecordQueryBuilder = $aqb;
  }

  /**
   * Includes access records for query access.
   *
   * @param \Drupal\entity\QueryAccess\QueryAccessEvent $event
   *   The event.
   */
  public function onQueryAccess(QueryAccessEvent $event) {
    $conditions = $event->getConditions();

    $account = $event->getAccount();
    $subject_type_id = 'user';
    $subject_id = $account->id();
    if (!($subject = $this->entityTypeManager->getStorage($subject_type_id)->load($subject_id))) {
      // Without a corresponding entity, we cannot have matching records.
      // @todo Clarify whether this goes too far.
      $conditions->addCacheContexts(['user']);
      $conditions->mergeCacheMaxAge(0);
      return;
    }

    $target_type_id = $event->getEntityTypeId();
    $operation = $event->getOperation();

    $ar_types = AccessRecordType::loadForAccessCheck($subject, $target_type_id, $operation, $conditions);
    if (empty($ar_types)) {
      return;
    }

    $ar_conditions = new ConditionGroup('OR');
    /** @var \Drupal\access_records\AccessRecordTypeInterface $ar_type */
    foreach ($ar_types as $ar_type) {
      if ($ar_condition = $this->buildAccessRecordCondition($ar_type, $operation, $subject_id)) {
        $ar_conditions->addCondition($ar_condition->getField(), $ar_condition->getValue(), $ar_condition->getOperator());
      }
    }

    if ($ar_conditions->count()) {
      $conditions->alwaysFalse(FALSE);
      $conditions->addCondition($ar_conditions);
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    return ['entity.query_access' => 'onQueryAccess'];
  }

  /**
   * Builds up a condition for access records of given type and operation.
   *
   * @param \Drupal\access_records\AccessRecordTypeInterface $ar_type
   *   The access record type.
   * @param string $operation
   *   The operation ("view", "update" or "delete").
   * @param int $subject_id
   *   The subject ID.
   *
   * @return \Drupal\entity\QueryAccess\Condition|null
   *   The condition, or NULL if no condition was built.
   */
  protected function buildAccessRecordCondition(AccessRecordTypeInterface $ar_type, string $operation, int $subject_id): ?Condition {
    if (!($id_query = $this->accessRecordQueryBuilder->selectByType($ar_type, $subject_id, $operation))) {
      return NULL;
    }

    $target_id_query = \Drupal::database()->select($id_query, 'ids');
    $target_id_query->addField('ids', 'target_id', 'target_id');

    return new Condition($ar_type->getTargetType()->getKey('id'), $target_id_query, 'IN');
  }

}
